// SPDX-License-Identifier: Unlicense
pragma solidity ^0.8.24;

import "./DAO.sol";
import "./ReputationHolder.sol";

struct AvailabilityStake {
    address worker;
    uint256 tokenId;
    uint256 amount;
    bool assigned;
}

enum WorkStatus {
    Requested,
    EvidenceSubmitted,
    ApprovalSubmitted,
    Complete
}

struct WorkRequest {
    address customer;
    uint256 fee;
    WorkStatus status;
    uint stakeIndex;
    bool approval;
    uint poolIndex;
}

contract Work1 is ReputationHolder {
    DAO immutable dao;
    uint public immutable price;
    mapping(uint => AvailabilityStake) stakes;
    uint stakeCount;
    mapping(uint => WorkRequest) public requests;
    uint public requestCount;

    // TODO: Make parameters configurable
    uint constant POOL_DURATION = 1 days;

    event WorkAssigned(address worker, uint requestIndex);

    constructor(DAO dao_, uint price_) {
        dao = dao_;
        price = price_;
    }

    /// Accept availability stakes as reputation token transfer
    function onERC721Received(
        address,
        address from,
        uint256 tokenId,
        bytes calldata
    ) public override returns (bytes4) {
        AvailabilityStake storage stake = stakes[stakeCount++];
        stake.worker = from;
        stake.tokenId = tokenId;
        stake.amount = dao.verifiedValueOf(from, tokenId);
        // TODO: use `data` parameter to include stake options such as duration
        return super.onERC721Received.selector;
    }

    /// Select a worker randomly from among the available workers, weighted by amount staked
    function randomWeightedSelection() internal view returns (uint) {
        uint totalStakes;
        for (uint i = 0; i < stakeCount; i++) {
            if (stakes[i].assigned) continue;
            totalStakes += stakes[i].amount;
        }
        uint select = block.prevrandao % totalStakes;
        uint acc;
        for (uint i = 0; i < stakeCount; i++) {
            if (stakes[i].assigned) continue;
            acc += stakes[i].amount;
            if (acc > select) return i;
        }
        revert("Failed to select worker");
    }

    /// Assign a random available worker
    function assignWork(uint requestIndex) internal returns (uint stakeIndex) {
        stakeIndex = randomWeightedSelection();
        AvailabilityStake storage stake = stakes[stakeIndex];
        stake.assigned = true;
        emit WorkAssigned(stake.worker, requestIndex);
    }

    /// Accept work request with fee
    function requestWork() external payable {
        require(msg.value >= price, "Insufficient fee");
        uint requestIndex = requestCount++;
        WorkRequest storage request = requests[requestIndex];
        request.customer = msg.sender;
        request.fee = msg.value;
        request.stakeIndex = assignWork(requestIndex);
    }

    /// Accept work evidence from worker
    function submitWorkEvidence(uint requestIndex) external {
        WorkRequest storage request = requests[requestIndex];
        require(request.status == WorkStatus.Requested);
        AvailabilityStake storage stake = stakes[request.stakeIndex];
        require(stake.worker == msg.sender);
        request.status = WorkStatus.EvidenceSubmitted;
    }

    /// Accept work approval/disapproval from customer
    function submitWorkApproval(uint requestIndex, bool approval) external {
        WorkRequest storage request = requests[requestIndex];
        require(request.status == WorkStatus.EvidenceSubmitted);
        AvailabilityStake storage stake = stakes[request.stakeIndex];
        request.status = WorkStatus.ApprovalSubmitted;
        request.approval = approval;
        // Initiate validation pool
        request.poolIndex = dao.initiateValidationPool(
            stake.worker,
            POOL_DURATION
        );
    }
}
